""" Monkey patches for loopy.symbolic

Use this module to insert pymbolic nodes and the likes.
"""
from dune.perftool.error import PerftoolError
from pymbolic.mapper.substitutor import make_subst_func

import loopy as lp
import pymbolic.primitives as prim


#
# Pymbolic nodes to insert into the symbolic language understood by loopy
#


class SumfactKernel(prim.Variable):
    def __init__(self,
                 a_matrices,
                 buffer,
                 stage,
                 preferred_position,
                 restriction,
                 ):
        self.a_matrices = a_matrices
        self.buffer = buffer
        self.stage = stage
        self.preferred_position = preferred_position
        self.restriction = restriction

        prim.Variable.__init__(self, "SUMFACT")

    def __getinitargs__(self):
        return (self.a_matrices, self.buffer, self.stage, self.preferred_position, self.restriction)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    init_arg_names = ("a_matrices", "buffer", "stage", "preferred_position", "restriction")

    mapper_method = "map_sumfact_kernel"


#
# Mapper methods to monkey patch into the visitor base classes!
#


def identity_map_sumfact_kernel(self, expr, *args):
    return expr


def walk_map_sumfact_kernel(self, expr, *args):
    self.visit(expr)


def stringify_map_sumfact_kernel(self, expr, *args):
    return "SUMFACT"


def dependency_map_sumfact_kernel(self, expr):
    return set()


def needs_resolution(self, expr):
    raise PerftoolError("SumfactKernel node is a placeholder and needs to be removed!")


#
# Do the actual monkey patching!!!
#


lp.symbolic.IdentityMapper.map_sumfact_kernel = identity_map_sumfact_kernel
lp.symbolic.SubstitutionMapper.map_sumfact_kernel = lp.symbolic.SubstitutionMapper.map_variable
lp.symbolic.WalkMapper.map_sumfact_kernel = walk_map_sumfact_kernel
lp.symbolic.StringifyMapper.map_sumfact_kernel = stringify_map_sumfact_kernel
lp.symbolic.DependencyMapper.map_sumfact_kernel = dependency_map_sumfact_kernel
lp.target.c.codegen.expression.ExpressionToCExpressionMapper.map_sumfact_kernel = needs_resolution
lp.type_inference.TypeInferenceMapper.map_sumfact_kernel = needs_resolution


#
# Some helper functions!
#


def substitute(expr, replacemap):
    """ A replacement for pymbolic.mapper.subsitutor.substitute which is aware of all
    monkey patches etc.
    """
    return lp.symbolic.SubstitutionMapper(make_subst_func(replacemap))(expr)
